//
//  UNST2.CPP
//
//  Michael Schindler
//  September 25, 1996
//  http://eiunix.tuwien.ac.at/~michael
//
// DESCRIPTION
// -----------
//
//  This program performs a reverse S transform with order 2 context on
//  an input file or stream, and sends the result to an output file or stream.
//
//  This program takes two arguments: an input file and an output
//  file.  You can leave off one argument and send your output to
//  stdout.  Leave off two arguments and read your input from stdin
//  as well.
//
//  The ST input looks like this:
//
//  long blocksize   { start-char  2nd-char  ...data... }*
//
//  This program accompanies Mark Nelsons programs to his article
//  "Data Compression with the Burrows-Wheeler Transform."
//
//
// Build Instructions
// ------------------
//
//  Define the constant unix for UNIX or UNIX-like systems.  The
//  use of this constant turns off the code used to force the MS-DOS
//  file system into binary mode.  g++ already does this, your UNIX
//  C++ compiler might also.
//
//  Borland C++ 4.5 16 bit    : bcc -w -ml unst2.cpp
//  Borland C++ 4.5 32 bit    : bcc32 -w unst2.cpp
//  Microsoft Visual C++ 1.5  : cl /W4 /AL unst2.cpp
//  Microsoft Visual C++ 4.0  : cl /W4 unst2.cpp
//  g++                       : g++ -o unst2 unst2.cpp
//
// Typical Use
// -----------
//
//  unari < raw-file | unrle | unmtf | unst2 > raw-file
//

#include <stdio.h>
#if !defined( unix )
#include <io.h>
#endif
#include <string.h>

#if ( INT_MAX == 32767 )
#define SIZE_SHIFT 14
#else
#define SIZE_SHIFT 18
#endif
#define BLOCK_SIZE (1 << SIZE_SHIFT)

unsigned char inbuffer[ BLOCK_SIZE ];
unsigned char outbuffer[ BLOCK_SIZE ];
unsigned int blocksize, buflen;
unsigned int count[ 256 ][ 256 ];

main( int argc, char *argv[] )
{
    int i, j, sum;
    fprintf( stderr, "Performing UNST2 on " );
    if ( argc > 1 ) {
        freopen( argv[ 1 ], "rb", stdin );
        fprintf( stderr, "%s", argv[ 1 ] );
    } else
        fprintf( stderr, "stdin" );
    fprintf( stderr, " to " );
    if ( argc > 2 ) {
        freopen( argv[ 2 ], "wb", stdout );
        fprintf( stderr, "%s", argv[ 2 ] );
    } else
        fprintf( stderr, "stdout" );
    fprintf( stderr, "\n" );
#if !defined( unix )
    setmode( fileno( stdin ), O_BINARY );
    setmode( fileno( stdout ), O_BINARY );
#endif
    if ( (blocksize = getc(stdin)) < 0 ) {
        fprintf( stderr, "Error reading data\n" );
        abort();
    }
    blocksize = 1 << blocksize;
    if ( blocksize > BLOCK_SIZE ) {
        fprintf( stderr, "Buffer overflow\n" );
        abort();
    }
    for ( ; ; ) {
        if ( (i = getc(stdin)) < 0 )
            break;
        outbuffer[ 0 ] = i;
        outbuffer[ 1 ] = getc(stdin);
        buflen = fread( (char *)inbuffer, 1, (size_t)blocksize, stdin);
        fprintf( stderr, "Processing %ld bytes\n", buflen );

        for ( i=0; i<256; i++ )
            for ( j=0; j < 256; j++ )
                count[ i ][ j ] = 0;                    // set counters to zero
        for ( i=0; i<buflen; i++ )
            count[ 0 ][ inbuffer[ i ] ]++;              // count frequencies
        sum = 0;
        for ( i=0; i<256; i++ ) {
            j = sum;
            sum += count[ 0 ][ i ];
            count[ 0 ][ i ] = 0;
            for ( ; j<sum; j++)
              count[ inbuffer[j] ][ i ]++;              // count contexts
        }
        if (j != buflen)
          fprintf(stdout,"Err2: j=%d b=%d\n",j, buflen);
        sum = 0;
        for ( i=0; i<256; i++ )
            for ( j=0; j < 256; j++ ) {
                sum += count[ i ][ j ];
                count[ i ][ j ] = sum - count[ i ][ j ];
            }
        if (sum != buflen)
          fprintf(stdout,"Err2: s=%d b=%d\n",sum, buflen);
        for ( i=2; i<buflen; i++ )
             outbuffer[ i ] =
                          inbuffer[ count[outbuffer[i-1]][outbuffer[i-2]]++ ];

        fwrite( (char *)outbuffer, 1, (size_t)buflen, stdout );
    }
    return 1;
}

